@@ -123,10 +123,13 @@ typedef enum _ZsOpt
     SOPT_FOCUS_FIT_WINDOW,
     SOPT_ALLWAYS_FOCUS_FIT_WINDOW,
     SOPT_SCALE_MOUSE,
+    SOPT_SCALE_MOUSE_DYNAMIC,
+    SOPT_SCALE_MOUSE_STATIC,
     SOPT_HIDE_ORIGINAL_MOUSE,
     SOPT_RESTRAIN_MOUSE,
     SOPT_RESTRAIN_MARGIN,
     SOPT_MOUSE_PAN,
+    SOPT_MINIMUM_ZOOM,
     SOPT_NUM
 } ZoomScreenOptions;
 
@@ -783,13 +786,26 @@ panZoom (CompScreen *s, int xvalue, int yvalue)
     constrainZoomTranslate (s);
 }
 
-/* Sets the zoom (or scale) level. Takes two float's as argument for legacy
- * reasons mainly; it uses the larger of the two, allowing us to easily
- * scale to fit an area. FIXME: This shouldn't happen here. */
+/* Enables polling of mouse position, and refreshes currently
+ * stored values.
+ */
 static void
-setScale (CompScreen *s, int out, float x, float y)
+enableMousePolling (CompScreen *s)
+{
+    ZOOM_SCREEN (s);
+    ZOOM_DISPLAY (s->display);
+    zs->pollHandle = 
+	(*zd->mpFunc->addPositionPolling) (s, updateMouseInterval);
+    zs->lastChange = time(NULL);
+    (*zd->mpFunc->getCurrentPosition) (s, &zs->mouseX, &zs->mouseY);
+}
+
+/* Sets the zoom (or scale) level. 
+ * Cleans up if we are suddenly zoomed out. 
+ */
+static void
+setScale (CompScreen *s, int out, float value)
 {
-    float value = x > y ? x : y;
     ZOOM_SCREEN(s);
 
     if (zs->zooms[out].locked)
@@ -800,11 +816,7 @@ setScale (CompScreen *s, int out, float x, float y)
     else
     {
 	if (!zs->pollHandle)
-	{
-	    ZOOM_DISPLAY (s->display);
-	    zs->pollHandle = 
-		(*zd->mpFunc->addPositionPolling) (s, updateMouseInterval);
-	}
+	    enableMousePolling (s);
 	zs->grabbed |= (1 << zs->zooms[out].output);
 	cursorZoomActive (s);
     }
@@ -816,10 +828,22 @@ setScale (CompScreen *s, int out, float x, float y)
 	cursorZoomInactive (s);
     }
 
+    if (value < zs->opt[SOPT_MINIMUM_ZOOM].value.f)
+	value = zs->opt[SOPT_MINIMUM_ZOOM].value.f;
+
     zs->zooms[out].newZoom = value;
     damageScreen(s);
 }
 
+/* Sets the zoom factor to the bigger of the two floats supplied. 
+ * Convenince function for setting the scale factor for an area.
+ */
+static inline void
+setScaleBigger (CompScreen *s, int out, float x, float y)
+{
+    setScale (s, out, x > y ? x : y);
+}
+
 /* Mouse code...
  * This takes care of keeping the mouse in sync with the zoomed area and
  * vice versa. 
@@ -1235,6 +1259,7 @@ drawCursor (CompScreen          *s,
     if (zs->cursor.isSet)
     {
 	CompTransform sTransform = *transform;
+	float	      scaleFactor;
 	int           ax, ay, x, y;
 	
 	/* This is a hack because these transformations are wrong when
@@ -1253,8 +1278,12 @@ drawCursor (CompScreen          *s,
         glPushMatrix ();
 	glLoadMatrixf (sTransform.m);
 	glTranslatef ((float) ax, (float) ay, 0.0f);
-	glScalef (1.0f / zs->zooms[out].currentZoom,
-		  1.0f / zs->zooms[out].currentZoom,
+	if (zs->opt[SOPT_SCALE_MOUSE_DYNAMIC].value.b) 
+	    scaleFactor = 1.0f / zs->zooms[out].currentZoom;
+	else 
+	    scaleFactor = 1.0f / zs->opt[SOPT_SCALE_MOUSE_STATIC].value.f;
+	glScalef (scaleFactor,
+		  scaleFactor,
 		  1.0f);
 	x = -zs->cursor.hotX;
 	y = -zs->cursor.hotY;
@@ -1475,8 +1504,8 @@ setZoomAreaAction (CompDisplay     *d,
 	setZoomArea (s, x1, y1, WIDTH, HEIGHT, FALSE);
 	o = &s->outputDev[out];
 	if (scale && WIDTH && HEIGHT)
-	    setScale (s, out, (float) WIDTH/o->width, 
-		      (float) HEIGHT/o->height);
+	    setScaleBigger (s, out, (float) WIDTH/o->width, 
+			    (float) HEIGHT/o->height);
 #undef WIDTH
 #undef HEIGHT
 	if (restrain)
@@ -1533,8 +1562,8 @@ ensureVisibilityAction (CompDisplay     *d,
 #define WIDTH (x2 - x1)
 #define HEIGHT (y2 - y1)
 	if (scale && WIDTH && HEIGHT)
-	    setScale (s, out, (float) WIDTH/o->width, 
-		      (float) HEIGHT/o->height);
+	    setScaleBigger (s, out, (float) WIDTH/o->width, 
+			    (float) HEIGHT/o->height);
 #undef WIDTH
 #undef HEIGHT
 	if (restrain)
@@ -1606,7 +1635,8 @@ zoomBoxDeactivate (CompDisplay     *d,
 
 	    out = outputDeviceForGeometry (s, x,y,width,height,0);
 	    o = &s->outputDev[out];
-	    setScale (s, out, (float) width/o->width, (float)  height/o->height);
+	    setScaleBigger (s, out, (float) width/o->width, (float)
+			    height/o->height);
 	    setZoomArea (s, x,y,width,height,FALSE);
 	}
     }
@@ -1637,8 +1667,8 @@ zoomIn (CompDisplay     *d,
 	    setCenter (s, pointerX, pointerY, TRUE);
 
 	setScale (s, out,
-		  zs->zooms[out].newZoom/zs->opt[SOPT_ZOOM_FACTOR].value.f,
-		  -1.0f);
+		  zs->zooms[out].newZoom /
+		  zs->opt[SOPT_ZOOM_FACTOR].value.f);
     }
     return TRUE;
 }
@@ -1701,7 +1731,7 @@ zoomSpecific (CompDisplay     *d,
 	if (otherScreenGrabExist (s, 0))
 	    return FALSE;
 
-	setScale (s, out, target, target);
+	setScale (s, out, target);
 
 	w = findWindowAtDisplay(d, d->activeWindow);
 	if (zd->opt[DOPT_SPECIFIC_TARGET_FOCUS].value.b
@@ -1788,7 +1818,8 @@ zoomToWindow (CompDisplay     *d,
     height = w->height + w->input.top + w->input.bottom;
     out = outputDeviceForWindow (w);
     o = &s->outputDev[out];
-    setScale (s, out, (float) width/o->width, (float)  height/o->height);
+    setScaleBigger (s, out, (float) width/o->width, 
+		    (float) height/o->height);
     zoomAreaToWindow (w);
     return TRUE;
 }
@@ -1990,8 +2021,7 @@ zoomOut (CompDisplay     *d,
 
 	setScale (s, out,
 		  zs->zooms[out].newZoom *
-		  zs->opt[SOPT_ZOOM_FACTOR].value.f,
-		  -1.0f);
+		  zs->opt[SOPT_ZOOM_FACTOR].value.f);
     }
 
     return TRUE;
@@ -2079,9 +2109,9 @@ focusTrack (CompDisplay *d,
 	int width = w->width + w->input.left + w->input.right;
 	int height = w->height + w->input.top + w->input.bottom;
 	
-	setScale (w->screen, out,
-		  (float) width / w->screen->outputDev[out].width,
-		  (float)  height/w->screen->outputDev[out].height);
+	setScaleBigger (w->screen, out,
+			(float) width / w->screen->outputDev[out].width,
+			(float)  height/w->screen->outputDev[out].height);
     }
     zoomAreaToWindow (w);
 }
@@ -2173,10 +2203,13 @@ static const CompMetadataOptionInfo zoomScreenOptionInfo[] = {
     { "focus_fit_window", "bool", "<default>false</default>", 0, 0 },
     { "always_focus_fit_window", "bool", "<default>false</default>", 0, 0 },
     { "scale_mouse", "bool", "<default>false</default>", 0, 0 },
+    { "scale_mouse_dynamic", "bool", "<default>true</default>", 0, 0 },
+    { "scale_mouse_static", "float", "<default>0.8</default>", 0, 0 },
     { "hide_original_mouse", "bool", "<default>false</default>", 0, 0 },
     { "restrain_mouse", "bool", "<default>false</default>", 0, 0 },
     { "restrain_margin", "int", "<default>5</default>", 0, 0 },
-    { "mouse_pan", "bool", "<default>false</default>", 0, 0 }
+    { "mouse_pan", "bool", "<default>false</default>", 0, 0 },
+    { "minimum_zoom", "float", "<max>1.00</max>", 0, 0 }
 };
 
 static CompOption *
@@ -2334,6 +2367,7 @@ zoomInitScreen (CompPlugin *p,
 	    break;
 	initialiseZoomArea (&zs->zooms[i], i);
     }
+    zs->lastChange = 0;
     zs->grabbed = 0;
     zs->mouseX = -1;
     zs->mouseY = -1;
@@ -2426,9 +2460,11 @@ zoomGetObjectOptions (CompPlugin *plugin,
 	(GetPluginObjectOptionsProc) zoomGetDisplayOptions,
 	(GetPluginObjectOptionsProc) zoomGetScreenOptions
     };
+    
+    *count = 0;
 
     RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab),
-		     (void *) (*count = 0), (plugin, object, count));
+		     NULL, (plugin, object, count));
 }
 
 static CompBool
