@@ -1,6 +1,25 @@
-
-/* Copyright Robert Carr 2007 blablabla GPL etc
- * racarr@beryl-project.org */
+/**
+ *
+ * Compiz metacity like info during resize
+ *
+ * resizeinfo.c
+ *
+ * Copyright (c) 2007 Robert Carr <racarr@opencompositing.org>
+ *
+ * Compiz resize atom usage and general cleanups by
+ * Copyright (c) 2007 Danny Baumann <maniac@opencompositing.org>
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License
+ * as published by the Free Software Foundation; either version 2
+ * of the License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ **/
 
 #define _GNU_SOURCE
 
@@ -8,7 +27,7 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include <compiz.h>
+#include <compiz-core.h>
 #include <cairo-xlib-xrender.h>
 #include <math.h>
 #include <pango/pango.h>
@@ -20,7 +39,7 @@
 
 static int displayPrivateIndex;
 
-typedef struct _InfoDisplay 
+typedef struct _InfoDisplay
 {
     int screenPrivateIndex;
 
@@ -46,32 +65,51 @@ typedef struct _InfoScreen
     DonePaintScreenProc    donePaintScreen;
   
     CompWindow *pWindow;
-  
+
     Bool drawing;
     int  fadeTime;
 
     InfoLayer backgroundLayer;
     InfoLayer textLayer;
-  
+ 
     XRectangle resizeGeometry;
 } InfoScreen;
 
-#define RESIZE_POPUP_WIDTH 75
+#define RESIZE_POPUP_WIDTH 85
 #define RESIZE_POPUP_HEIGHT 50
 
 #define GET_INFO_DISPLAY(d)				    \
-    ((InfoDisplay *) (d)->privates[displayPrivateIndex].ptr)
+    ((InfoDisplay *) (d)->base.privates[displayPrivateIndex].ptr)
 
 #define INFO_DISPLAY(d)			  \
     InfoDisplay *id = GET_INFO_DISPLAY (d)
 
 #define GET_INFO_SCREEN(s, id)					\
-    ((InfoScreen *) (s)->privates[(id)->screenPrivateIndex].ptr)
+    ((InfoScreen *) (s)->base.privates[(id)->screenPrivateIndex].ptr)
 
 #define INFO_SCREEN(s)						       \
     InfoScreen *is = GET_INFO_SCREEN (s, GET_INFO_DISPLAY (s->display))
 
 /* Set up an InfoLayer to build a cairo->opengl texture pipeline */
+static void
+freeInfoLayer (CompScreen *s,
+	       InfoLayer  *il)
+{
+    if (il->cr)
+	cairo_destroy (il->cr);
+    il->cr = NULL;
+
+    if (il->surface)
+	cairo_surface_destroy (il->surface);
+    il->surface = NULL;
+
+    finiTexture (s, &il->texture);
+
+    if (il->pixmap)
+	XFreePixmap (s->display->display, il->pixmap);
+    il->pixmap = None;
+}
+
 static void
 setupCairoLayer (CompScreen *s,
 		 InfoLayer  *il)
@@ -81,7 +119,11 @@ setupCairoLayer (CompScreen *s,
     int               w, h;
 	
     screen = ScreenOfDisplay (s->display->display, s->screenNum);
-	
+
+    memset (il, 0, sizeof (InfoLayer));
+
+    initTexture (s, &il->texture);
+
     w = RESIZE_POPUP_WIDTH;
     h = RESIZE_POPUP_HEIGHT;
 	
@@ -91,9 +133,9 @@ setupCairoLayer (CompScreen *s,
     il->pixmap = XCreatePixmap (s->display->display, s->root, w, h, 32);
     if (!bindPixmapToTexture (s, &il->texture, il->pixmap, w, h, 32))
     {
-	compLogMessage (s->display, "resizeinfo", CompLogLevelWarn,
+	compLogMessage ("resizeinfo", CompLogLevelWarn,
 			"Bind Pixmap to Texture failure");
-	XFreePixmap (s->display->display, il->pixmap);
+	freeInfoLayer (s, il);
 	return;
     }
 	
@@ -101,15 +143,20 @@ setupCairoLayer (CompScreen *s,
 	cairo_xlib_surface_create_with_xrender_format (s->display->display,
 						       il->pixmap, screen,
 						       format, w, h);
+    if (cairo_surface_status (il->surface) != CAIRO_STATUS_SUCCESS)
+    {
+	compLogMessage ("resizeinfo", CompLogLevelWarn,
+			"Could not create cairo layer surface,");
+	freeInfoLayer (s, il);
+	return;
+    }
+
     il->cr = cairo_create (il->surface);
     if (cairo_status (il->cr) != CAIRO_STATUS_SUCCESS)
     {
-	compLogMessage (s->display, "resizeinfo", CompLogLevelWarn,
+	compLogMessage ("resizeinfo", CompLogLevelWarn,
 			"Could not create cairo context");
-	cairo_destroy (il->cr);
-	cairo_surface_destroy (il->surface);
-	XFreePixmap (s->display->display, il->pixmap);
-	il->cr = NULL;
+	freeInfoLayer (s, il);
 	return;
     }
 }
@@ -125,7 +172,7 @@ updateTextLayer (CompScreen *s)
     int                  widthInc, heightInc;
     int                  width, height, xv, yv;
     unsigned short       *color;
-    char                 *info;
+    char                 info[50];
     cairo_t              *cr;
     PangoLayout          *layout;
     PangoFontDescription *font;
@@ -157,8 +204,8 @@ updateTextLayer (CompScreen *s)
     cairo_restore (cr);
     cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
 
-    asprintf (&info, "%d x %d", xv, yv);
- 
+    snprintf (info, 50, "%d x %d", xv, yv);
+
     font = pango_font_description_new ();
     layout = pango_cairo_create_layout (is->textLayer.cr);
   
@@ -269,6 +316,33 @@ gradientChanged (CompDisplay              *d,
 	drawCairoBackground (s);
 }
 
+static void
+damagePaintRegion (CompScreen *s)
+{
+    REGION reg;
+    int    x, y;
+
+    INFO_SCREEN (s);
+
+    if (!is->fadeTime && !is->drawing)
+	return;
+
+    x = is->resizeGeometry.x + is->resizeGeometry.width / 2.0f -
+	RESIZE_POPUP_WIDTH / 2.0f;
+    y = is->resizeGeometry.y + is->resizeGeometry.height / 2.0f - 
+	RESIZE_POPUP_HEIGHT / 2.0f;
+
+    reg.rects    = &reg.extents;
+    reg.numRects = 1;
+
+    reg.extents.x1 = x - 5;
+    reg.extents.y1 = y - 5;
+    reg.extents.x2 = x + RESIZE_POPUP_WIDTH + 5;
+    reg.extents.y2 = y + RESIZE_POPUP_HEIGHT + 5;
+
+    damageScreenRegion (s, &reg);
+}
+
 /*  Handle the fade in /fade out. */
 static void
 infoPreparePaintScreen (CompScreen *s,
@@ -282,7 +356,7 @@ infoPreparePaintScreen (CompScreen *s,
 	if (is->fadeTime < 0)
 	    is->fadeTime = 0;
     }
-	
+
     UNWRAP (is, s, preparePaintScreen);
     (*s->preparePaintScreen) (s, ms);
     WRAP (is, s, preparePaintScreen, infoPreparePaintScreen);
@@ -295,26 +369,8 @@ infoDonePaintScreen (CompScreen *s)
 
     if (is->pWindow)
     {
-	if (is->fadeTime || is->drawing)
-	{
-	    REGION reg;
-	    int    x, y;
-
-	    x = is->resizeGeometry.x + is->resizeGeometry.width / 2.0f -
-		RESIZE_POPUP_WIDTH / 2.0f;
-    	    y = is->resizeGeometry.y + is->resizeGeometry.height / 2.0f - 
-		RESIZE_POPUP_HEIGHT / 2.0f;
-
-	    reg.rects    = &reg.extents;
-	    reg.numRects = 1;
-
-	    reg.extents.x1 = x - 5;
-	    reg.extents.y1 = y - 5;
-	    reg.extents.x2 = x + RESIZE_POPUP_WIDTH + 5;
-	    reg.extents.y2 = y + RESIZE_POPUP_HEIGHT + 5;
-
-	    damageScreenRegion (s, &reg);
-	}
+	if (is->fadeTime)
+	    damagePaintRegion (s);
 	
 	if (!is->fadeTime && !is->drawing)
 	    is->pWindow = 0;
@@ -503,6 +559,7 @@ infoHandleEvent (CompDisplay *d,
 		    is->resizeGeometry.height = event->xclient.data.l[3];
 
 		    updateTextLayer (w->screen);
+		    damagePaintRegion (w->screen);
 		}
 	    }
 	}
@@ -522,6 +579,9 @@ infoInitDisplay (CompPlugin  *p,
 {
     InfoDisplay *id;
 
+    if (!checkPluginABI ("core", CORE_ABIVERSION))
+	return FALSE;
+
     id = malloc (sizeof (InfoDisplay));
     if (!id)
 	return FALSE;
@@ -539,7 +599,7 @@ infoInitDisplay (CompPlugin  *p,
 
     id->resizeNotifyAtom = XInternAtom (d->display, "_COMPIZ_RESIZE_NOTIFY", 0);
 
-    d->privates[displayPrivateIndex].ptr = id;
+    d->base.privates[displayPrivateIndex].ptr = id;
 
     WRAP (id, d, handleEvent, infoHandleEvent);
 
@@ -559,7 +619,7 @@ infoFiniDisplay (CompPlugin  *p,
     free (id);
 }
 
-static Bool 
+static Bool
 infoInitScreen (CompPlugin *p,
 		CompScreen *s)
 {
@@ -580,40 +640,22 @@ infoInitScreen (CompPlugin *p,
     is->resizeGeometry.width  = 0;
     is->resizeGeometry.height = 0;
 
-    initTexture (s, &is->backgroundLayer.texture);
-    initTexture (s, &is->textLayer.texture);
-	
     WRAP (is, s, windowGrabNotify, infoWindowGrabNotify);
     WRAP (is, s, windowUngrabNotify, infoWindowUngrabNotify);
     WRAP (is, s, preparePaintScreen, infoPreparePaintScreen);
     WRAP (is, s, paintOutput, infoPaintOutput);
     WRAP (is, s, donePaintScreen, infoDonePaintScreen);
 
-    s->privates[id->screenPrivateIndex].ptr = is;
+    s->base.privates[id->screenPrivateIndex].ptr = is;
 
     /* setup and draw cairo background */
     setupCairoLayer (s, &is->backgroundLayer);
     drawCairoBackground (s);
-	
+    
     /* setup text layer */
     setupCairoLayer (s, &is->textLayer);
-  
-    return TRUE;
-}
 
-static void
-freeInfoLayer (CompScreen *s,
-	       InfoLayer  *is)
-{
-    if (is->cr)
-	cairo_destroy (is->cr);
-    if (is->surface)
-	cairo_surface_destroy (is->surface);
-  
-    finiTexture (s, &is->texture);
-  
-    if (is->pixmap)
-	XFreePixmap (s->display->display, is->pixmap);
+    return TRUE;
 }
 
 static void
@@ -624,7 +666,7 @@ infoFiniScreen (CompPlugin *p,
 	
     freeInfoLayer (s, &is->backgroundLayer);
     freeInfoLayer (s, &is->textLayer);
-         
+
     UNWRAP (is, s, windowGrabNotify);
     UNWRAP (is, s, windowUngrabNotify);
     UNWRAP (is, s, preparePaintScreen);
@@ -634,6 +676,32 @@ infoFiniScreen (CompPlugin *p,
     free (is);
 }
 
+static CompBool
+infoInitObject (CompPlugin *p,
+		CompObject *o)
+{
+    static InitPluginObjectProc dispTab[] = {
+	(InitPluginObjectProc) 0, /* InitCore */
+	(InitPluginObjectProc) infoInitDisplay,
+	(InitPluginObjectProc) infoInitScreen
+    };
+
+    RETURN_DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), TRUE, (p, o));
+}
+
+static void
+infoFiniObject (CompPlugin *p,
+		CompObject *o)
+{
+    static FiniPluginObjectProc dispTab[] = {
+	(FiniPluginObjectProc) 0, /* FiniCore */
+	(FiniPluginObjectProc) infoFiniDisplay,
+	(FiniPluginObjectProc) infoFiniScreen
+    };
+
+    DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), (p, o));
+}
+
 static Bool
 infoInit (CompPlugin *p)
 {
@@ -650,29 +718,15 @@ infoFini (CompPlugin *p)
     freeDisplayPrivateIndex(displayPrivateIndex);
 }
 
-static int
-infoGetVersion(CompPlugin *plugin,
-	       int        version)
-{
-    return ABIVERSION;
-}
-
 static CompPluginVTable infoVTable = {
     "resizeinfo",
-    infoGetVersion,
     0,
     infoInit,
     infoFini,
-    infoInitDisplay,
-    infoFiniDisplay,
-    infoInitScreen,
-    infoFiniScreen,
-    0, /* InitWindow */
-    0, /* FiniWindow */
-    0, /* GetDisplayOptions */
-    0, /* SetDisplayOption */
-    0, /* GetScreenOptions */
-    0 /* SetScreenOption */
+    infoInitObject,
+    infoFiniObject,
+    0,
+    0
 };
 
 CompPluginVTable*
