@@ -94,7 +94,7 @@ typedef struct _SwitchScreen {
 #define BORDER 10
 
 #define SWITCH_DISPLAY(d) PLUGIN_DISPLAY(d, Switch, s)
-#define SWITCH_SCREEN(s) PLUGIN_SCREEN(s, Switch, s)
+#define SWITCH_SCREEN(screen) PLUGIN_SCREEN(screen, Switch, s)
 
 static void
 setSelectedWindowHint (CompScreen *s)
@@ -145,6 +145,9 @@ isSwitchWin (CompWindow *w)
 
 	if (w->state & CompWindowStateSkipTaskbarMask)
 	    return FALSE;
+
+	if (!matchEval (staticswitcherGetWindowMatch (s), w))
+	    return FALSE;
     }
 
     if (ss->selection == CurrentViewport)
@@ -170,9 +173,6 @@ isSwitchWin (CompWindow *w)
 	    return FALSE;
     }
 
-    if (!matchEval (staticswitcherGetWindowMatch (s), w))
-	return FALSE;
-
     return TRUE;
 }
 
@@ -280,18 +280,20 @@ switchUpdatePopupWindow (CompScreen *s,
     ss->previewHeight = h;
     ss->previewBorder = b;
 
+    x = s->outputDev[s->currentOutputDev].region.extents.x1 +
+	s->outputDev[s->currentOutputDev].width / 2;
+    y = s->outputDev[s->currentOutputDev].region.extents.y1 +
+	s->outputDev[s->currentOutputDev].height / 2;
+
     xsh.flags       = PSize | PPosition | PWinGravity;
+    xsh.x           = x;
+    xsh.y           = y;
     xsh.width       = winWidth;
     xsh.height      = winHeight;
     xsh.win_gravity = StaticGravity;
 
     XSetWMNormalHints (s->display->display, ss->popupWindow, &xsh);
 
-    x = s->outputDev[s->currentOutputDev].region.extents.x1 +
-	s->outputDev[s->currentOutputDev].width / 2;
-    y = s->outputDev[s->currentOutputDev].region.extents.y1 +
-	s->outputDev[s->currentOutputDev].height / 2;
-
     XMoveResizeWindow (s->display->display, ss->popupWindow,
 		       x - winWidth / 2, y - winHeight / 2,
 		       winWidth, winHeight);
@@ -333,6 +335,72 @@ switchCreateWindowList (CompScreen *s,
     switchUpdateWindowList (s, count);
 }
 
+static Bool
+switchGetPaintRectangle (CompWindow *w,
+			 BoxPtr     rect,
+			 int        *opacity)
+{
+    StaticswitcherHighlightRectHiddenEnum mode;
+
+    mode = staticswitcherGetHighlightRectHidden (w->screen);
+
+    if (w->attrib.map_state == IsViewable || w->shaded)
+    {
+	rect->x1 = w->attrib.x - w->input.left;
+	rect->y1 = w->attrib.y - w->input.top;
+	rect->x2 = w->attrib.x + w->width + w->input.right;
+	rect->y2 = w->attrib.y + w->height + w->input.bottom;
+	return TRUE;
+    }
+    else if (mode == HighlightRectHiddenTaskbarEntry && w->iconGeometrySet)
+    {
+	rect->x1 = w->iconGeometry.x;
+	rect->y1 = w->iconGeometry.y;
+	rect->x2 = rect->x1 + w->iconGeometry.width;
+	rect->y2 = rect->y1 + w->iconGeometry.height;
+	return TRUE;
+    }
+    else if (mode == HighlightRectHiddenOriginalWindowPosition)
+    {
+	rect->x1 = w->serverX - w->input.left;
+	rect->y1 = w->serverY - w->input.top;
+	rect->x2 = w->serverX + w->serverWidth + w->input.right;
+	rect->y2 = w->serverY + w->serverHeight + w->input.bottom;
+
+	if (opacity)
+	    *opacity /= 4;
+
+	return TRUE;
+    }
+
+    return FALSE;
+}
+
+static void
+switchDoWindowDamage (CompWindow *w)
+{
+    if (w->attrib.map_state == IsViewable || w->shaded)
+	addWindowDamage (w);
+    else
+    {
+	BoxRec box;
+	if (switchGetPaintRectangle (w, &box, NULL))
+	{
+	    REGION reg;
+
+	    reg.rects    = &reg.extents;
+	    reg.numRects = 1;
+
+	    reg.extents.x1 = box.x1 - 2;
+	    reg.extents.y1 = box.y1 - 2;
+	    reg.extents.x2 = box.x2 + 2;
+	    reg.extents.y2 = box.y2 + 2;
+
+	    damageScreenRegion (w->screen, &reg);
+	}
+    }
+}
+
 static void
 switchToWindow (CompScreen *s,
 		Bool	   toNext)
@@ -410,13 +478,13 @@ switchToWindow (CompScreen *s,
 	    setSelectedWindowHint (s);
 	}
 
-	addWindowDamage (w);
+	switchDoWindowDamage (w);
 
 	if (old)
 	{
 	    w = findWindowAtScreen (s, old);
 	    if (w)
-		addWindowDamage (w);
+		switchDoWindowDamage (w);
 	}
     }
 }
@@ -522,6 +590,7 @@ switchInitiate (CompScreen            *s,
     {
 	Display		     *dpy = s->display->display;
 	XWMHints	     xwmh;
+	XClassHint           xch;
 	Atom		     state[4];
 	int		     nState = 0;
 	XSetWindowAttributes attr;
@@ -534,6 +603,9 @@ switchInitiate (CompScreen            *s,
 	xwmh.flags = InputHint;
 	xwmh.input = 0;
 
+	xch.res_name  = "compiz";
+	xch.res_class = "switcher-window";
+
 	attr.background_pixel = 0;
 	attr.border_pixel     = 0;
 	attr.colormap	      = XCreateColormap (dpy, s->root, visual,
@@ -546,7 +618,7 @@ switchInitiate (CompScreen            *s,
 
 	XSetWMProperties (dpy, ss->popupWindow, NULL, NULL,
 			  programArgv, programArgc,
-			  NULL, &xwmh, NULL);
+			  NULL, &xwmh, &xch);
 
 	state[nState++] = s->display->winStateAboveAtom;
 	state[nState++] = s->display->winStateStickyAtom;
@@ -590,6 +662,7 @@ switchInitiate (CompScreen            *s,
 			compRemoveTimeout (ss->popupDelayHandle);
 
 		    ss->popupDelayHandle = compAddTimeout (delay,
+							   (float) delay * 1.2,
 							   switchShowPopup, s);
 		}
 		else
@@ -933,11 +1006,11 @@ switchWindowRemove (CompDisplay *d,
 
 	if (old != ss->selectedWindow)
 	{
-	    addWindowDamage (w);
+	    switchDoWindowDamage (w);
 
 	    w = findWindowAtScreen (w->screen, old);
 	    if (w)
-		addWindowDamage (w);
+		switchDoWindowDamage (w);
 
 	    ss->moreAdjust = 1;
 	}
@@ -964,7 +1037,7 @@ updateForegroundColor (CompScreen *s)
 				 XA_INTEGER, &actual, &format,
 				 &n, &left, &propData);
 
-    if (result == Success && n && propData)
+    if (result == Success && propData)
     {
 	if (n == 3 || n == 4)
 	{
@@ -996,6 +1069,24 @@ switchHandleEvent (CompDisplay *d,
     CompWindow *w;
     SWITCH_DISPLAY (d);
 
+    switch (event->type) {
+    case MapNotify:
+	w = findWindowAtDisplay (d, event->xmap.window);
+	if (w)
+	{
+	    SWITCH_SCREEN (w->screen);
+
+	    if (w->id == ss->popupWindow)
+	    {
+		w->wmType = getWindowType (d, w->id);
+		recalcWindowType (w);
+		recalcWindowActions (w);
+		updateWindowClassHints (w);
+	    }
+	}
+	break;
+    }
+
     UNWRAP (sd, d, handleEvent);
     (*d->handleEvent) (d, event);
     WRAP (sd, d, handleEvent, switchHandleEvent);
@@ -1095,6 +1186,21 @@ switchPreparePaintScreen (CompScreen *s,
     WRAP (ss, s, preparePaintScreen, switchPreparePaintScreen);
 }
 
+static inline void
+switchPaintRect (BoxRec *box,
+		 unsigned int offset,
+		 unsigned short *color,
+		 int opacity)
+{
+    glColor4us (color[0], color[1], color[2], color[3] * opacity / 100);
+    glBegin (GL_LINE_LOOP);
+    glVertex2i (box->x1 + offset, box->y1 + offset);
+    glVertex2i (box->x2 - offset, box->y1 + offset);
+    glVertex2i (box->x2 - offset, box->y2 - offset);
+    glVertex2i (box->x1 + offset, box->y2 - offset);
+    glEnd ();
+}
+
 static Bool
 switchPaintOutput (CompScreen		   *s,
 		   const ScreenPaintAttrib *sAttrib,
@@ -1109,10 +1215,10 @@ switchPaintOutput (CompScreen		   *s,
 
     if (ss->grabIndex)
     {
-	CompWindow    *zoomed;
-	CompWindow    *switcher;
-	Window	      zoomedAbove = None;
-	Bool	      saveDestroyed = FALSE;
+	StaticswitcherHighlightModeEnum mode;
+	CompWindow                      *switcher, *zoomed;
+	Window	                        zoomedAbove = None;
+	Bool	                        saveDestroyed = FALSE;
 
 	switcher = findWindowAtScreen (s, ss->popupWindow);
 	if (switcher)
@@ -1121,7 +1227,12 @@ switchPaintOutput (CompScreen		   *s,
 	    switcher->destroyed = TRUE;
 	}
 
-	if (staticswitcherGetBringToFront (s) && !ss->popupDelayHandle)
+	if (!ss->popupDelayHandle)
+	    mode = staticswitcherGetHighlightMode (s);
+	else
+	    mode = HighlightModeNone;
+
+	if (mode == HighlightModeBringSelectedToFront)
 	{
 	    zoomed = findWindowAtScreen (s, ss->selectedWindow);
 	    if (zoomed)
@@ -1151,23 +1262,74 @@ switchPaintOutput (CompScreen		   *s,
 	    insertWindowIntoScreen (s, zoomed, zoomedAbove);
 	}
 
-	if (switcher)
+	if (switcher || mode == HighlightModeShowRectangle)
 	{
 	    CompTransform sTransform = *transform;
 
-	    switcher->destroyed = saveDestroyed;
-
 	    transformToScreenSpace (s, output, -DEFAULT_Z_CAMERA, &sTransform);
 
 	    glPushMatrix ();
 	    glLoadMatrixf (sTransform.m);
 
-	    if (!switcher->destroyed			 &&
-		switcher->attrib.map_state == IsViewable &&
-		switcher->damaged)
+	    if (mode == HighlightModeShowRectangle)
 	    {
-		(*s->paintWindow) (switcher, &switcher->paint, &sTransform,
-				   &infiniteRegion, 0);
+		CompWindow *w;
+
+		if (zoomed)
+		    w = zoomed;
+		else
+		    w = findWindowAtScreen (s, ss->selectedWindow);
+
+		if (w)
+		{
+		    BoxRec box;
+		    int    opacity = 100;
+
+		    if (switchGetPaintRectangle (w, &box, &opacity))
+		    {
+			unsigned short *color;
+			GLushort       r, g, b, a;
+
+			glEnable (GL_BLEND);
+
+			/* fill rectangle */
+			r = staticswitcherGetHighlightColorRed (s);
+			g = staticswitcherGetHighlightColorGreen (s);
+			b = staticswitcherGetHighlightColorBlue (s);
+			a = staticswitcherGetHighlightColorAlpha (s);
+			a = a * opacity / 100;
+
+			glColor4us (r, g, b, a);
+			glRecti (box.x1, box.y2, box.x2, box.y1);
+
+			/* draw outline */
+			glLineWidth (1.0);
+			glDisable (GL_LINE_SMOOTH);
+
+			color = staticswitcherGetHighlightBorderColor (s);
+			switchPaintRect (&box, 0, color, opacity);
+			switchPaintRect (&box, 2, color, opacity);
+			color = staticswitcherGetHighlightBorderInlayColor (s);
+			switchPaintRect (&box, 1, color, opacity);
+
+			/* clean up */
+			glColor4usv (defaultColor);
+			glDisable (GL_BLEND);
+		    }
+		}
+	    }
+
+	    if (switcher)
+	    {
+		switcher->destroyed = saveDestroyed;
+
+		if (!switcher->destroyed		     &&
+		    switcher->attrib.map_state == IsViewable &&
+		    switcher->damaged)
+		{
+		    (*s->paintWindow) (switcher, &switcher->paint, &sTransform,
+				       &infiniteRegion, 0);
+		}
 	    }
 
 	    glPopMatrix ();
@@ -1463,14 +1625,25 @@ switchGetRowXOffset (CompScreen   *s,
 		     SwitchScreen *ss,
 		     int          y)
 {
-    if (staticswitcherGetRowAlign (s) == RowAlignLeft)
-	return 0;
+    int retval = 0;
 
     if (ss->nWindows - (y * ss->xCount) >= ss->xCount)
 	return 0;
 
-    return (ss->xCount - ss->nWindows + (y * ss->xCount)) *
-	   (ss->previewWidth + ss->previewBorder) / 2;
+    switch (staticswitcherGetRowAlign (s)) {
+    case RowAlignLeft:
+	break;
+    case RowAlignCentered:
+	retval = (ss->xCount - ss->nWindows + (y * ss->xCount)) *
+	         (ss->previewWidth + ss->previewBorder) / 2;
+	break;
+    case RowAlignRight:
+	retval = (ss->xCount - ss->nWindows + (y * ss->xCount)) *
+	         (ss->previewWidth + ss->previewBorder);
+	break;
+    }
+
+    return retval;
 }
 
 static Bool
