@@ -30,45 +30,32 @@
 #include <sys/types.h>
 #include <unistd.h>
 
-#include <X11/Xatom.h>
-#include <X11/extensions/Xrender.h>
+#include <compiz-core.h>
+#include <compiz-text.h>
+#include <compiz-mousepoll.h>
 
-#include <compiz.h>
-#include <text.h>
+#include <X11/Xatom.h>
 
 #include "thumbnail_tex.h"
 #include "thumbnail_options.h"
 
 #define GET_THUMB_DISPLAY(d)				       \
-    ((ThumbDisplay *) (d)->privates[displayPrivateIndex].ptr)
+    ((ThumbDisplay *) (d)->base.privates[displayPrivateIndex].ptr)
 
 #define THUMB_DISPLAY(d)		       \
     ThumbDisplay *td = GET_THUMB_DISPLAY (d)
 
 #define GET_THUMB_SCREEN(s, td)				   \
-    ((ThumbScreen *) (s)->privates[(td)->screenPrivateIndex].ptr)
+    ((ThumbScreen *) (s)->base.privates[(td)->screenPrivateIndex].ptr)
 
 #define THUMB_SCREEN(s)						      \
     ThumbScreen *ts = GET_THUMB_SCREEN (s, GET_THUMB_DISPLAY (s->display))
 
-#define GET_THUMB_WINDOW(w, ts)                               \
-    ((ThumbWindow *) (w)->privates[(ts)->windowPrivateIndex].ptr)
-
-#define THUMB_WINDOW_PTR(w)                                   \
-    GET_THUMB_WINDOW  (w,                                    \
-		       GET_THUMB_SCREEN  (w->screen,                            \
-					  GET_THUMB_DISPLAY (w->screen->display)))
-
-#define THUMB_WINDOW(w)                                       \
-    ThumbWindow *tw = THUMB_WINDOW_PTR(w)
-
 #define WIN_X(w) ((w)->attrib.x - (w)->input.left)
 #define WIN_Y(w) ((w)->attrib.y - (w)->input.top)
 #define WIN_W(w) ((w)->width + (w)->input.left + (w)->input.right)
 #define WIN_H(w) ((w)->height + (w)->input.top + (w)->input.bottom)
 
-#define THUMB_MOUSE_UPDATE_SPEED 100
-
 #define TEXT_DISTANCE 10
 
 static int displayPrivateIndex;
@@ -79,9 +66,10 @@ typedef struct _ThumbDisplay
 
     HandleEventProc handleEvent;
 
-    Atom winIconGeometryAtom;
-}
-ThumbDisplay;
+    Bool textAvailable;
+
+    MousePollFunc *mpFunc;
+} ThumbDisplay;
 
 typedef struct _Thumbnail
 {
@@ -98,17 +86,13 @@ typedef struct _Thumbnail
 
     CompTexture textTexture;
     Pixmap      textPixmap;
-    
+
     int tWidth;
     int tHeight;
-}
-Thumbnail;
+} Thumbnail;
 
 typedef struct _ThumbScreen
 {
-    int windowPrivateIndex;
-
-    CompTimeoutHandle mouseTimeout;
     CompTimeoutHandle displayTimeout;
 
     PreparePaintScreenProc     preparePaintScreen;
@@ -121,7 +105,7 @@ typedef struct _ThumbScreen
 
     CompWindow *dock;
     CompWindow *pointedWin;
-    
+
     Bool      showingThumb;
     Thumbnail thumb;
     Thumbnail oldThumb;
@@ -133,27 +117,8 @@ typedef struct _ThumbScreen
     int x;
     int y;
 
-}
-ThumbScreen;
-
-typedef struct _IconGeometry
-{
-    int x;
-    int y;
-    int width;
-    int height;
-    
-    Bool isSet;
-}
-
-IconGeometry;
-
-typedef struct _ThumbWindow
-{
-    IconGeometry ig;
-}
-
-ThumbWindow;
+    PositionPollingHandle pollHandle;
+} ThumbScreen;
 
 static void
 freeThumbText (CompScreen *s,
@@ -176,23 +141,27 @@ renderThumbText (CompScreen *s,
 		 Thumbnail  *t,
 		 Bool       freeThumb)
 {
-    int stride;
-    void *data;
+    int            stride;
+    void           *data;
+    CompTextAttrib tA;
+
+    THUMB_DISPLAY (s->display);
 
     if (freeThumb)
 	freeThumbText (s, t);
 
-    CompTextAttrib tA;
+    if (!td->textAvailable)
+	return;
 
-    tA.maxwidth   = t->width;
-    tA.maxheight  = 100;
+    tA.maxWidth   = t->width;
+    tA.maxHeight  = 100;
     tA.screen     = s;
     tA.size       = thumbnailGetFontSize (s);
     tA.color[0]   = thumbnailGetFontColorRed (s);
     tA.color[1]   = thumbnailGetFontColorGreen (s);
     tA.color[2]   = thumbnailGetFontColorBlue (s);
     tA.color[3]   = thumbnailGetFontColorAlpha (s);
-    tA.style      = (thumbnailGetFontBold (s) ) ?
+    tA.style      = (thumbnailGetFontBold (s)) ?
 		    TEXT_STYLE_BOLD : TEXT_STYLE_NORMAL;
     tA.family     = "Sans";
     tA.ellipsize  = TRUE;
@@ -201,8 +170,8 @@ renderThumbText (CompScreen *s,
 
     initTexture (s, &t->textTexture);
 
-    if ( (*s->display->fileToImage) (s->display, TEXT_ID, (char *) &tA,
-				     &t->tWidth, &t->tHeight, &stride, &data) )
+    if ((*s->display->fileToImage) (s->display, TEXT_ID, (char *) &tA,
+				    &t->tWidth, &t->tHeight, &stride, &data))
     {
 	t->textPixmap = (Pixmap) data;
 	bindPixmapToTexture (s, &t->textTexture, t->textPixmap,
@@ -216,44 +185,12 @@ renderThumbText (CompScreen *s,
     }
 }
 
-
-static void
-updateWindowIconGeometry (CompWindow *w)
-{
-    THUMB_DISPLAY (w->screen->display);
-    THUMB_WINDOW (w);
-
-    Atom actual;
-    int result, format;
-    unsigned long n, left;
-    unsigned char *data;
-    unsigned long *mydata;
-
-    result = XGetWindowProperty (w->screen->display->display,
-				 w->id, td->winIconGeometryAtom, 0L, 4L,
-				 FALSE, XA_CARDINAL, &actual, &format, &n,
-				 &left, &data);
-
-    mydata = (unsigned long *) data;
-
-    tw->ig.isSet = FALSE;
-
-    if (result == Success && actual == XA_CARDINAL && n == 4)
-    {
-	tw->ig.x      = mydata[0];
-	tw->ig.y      = mydata[1];
-	tw->ig.width  = mydata[2];
-	tw->ig.height = mydata[3];
-	tw->ig.isSet  = TRUE;
-	XFree (data);
-    }
-}
-
 static void
 damageThumbRegion (CompScreen *s,
 		   Thumbnail  *t)
 {
     REGION region;
+
     region.extents.x1 = t->x - t->offset;
     region.extents.y1 = t->y - t->offset;
     region.extents.x2 = region.extents.x1 + t->width + (t->offset * 2);
@@ -275,10 +212,16 @@ damageThumbRegion (CompScreen *s,
 static void
 thumbUpdateThumbnail (CompScreen *s)
 {
-    THUMB_SCREEN (s);
+    int        igMidPoint[2], tMidPoint[2];
+    int        tPos[2], tmpPos[2];
+    float      distance = 1000000;
+    int        off, oDev, tHeight;
+    int        ox1, oy1, ox2, oy2, ow, oh;
+    float      maxSize = thumbnailGetThumbSize (s);
+    double     scale  = 1.0;
+    CompWindow *w;
 
-    float maxSize = thumbnailGetThumbSize (s);
-    double scale  = 1.0;
+    THUMB_SCREEN (s);
 
     if (ts->thumb.win == ts->pointedWin)
 	return;
@@ -295,43 +238,42 @@ thumbUpdateThumbnail (CompScreen *s)
     ts->thumb.win  = ts->pointedWin;
     ts->thumb.dock = ts->dock;
 
-    if (!ts->thumb.win)
-	return;
-
-    /* do we nee to scale the window down? */
-    if (WIN_W (ts->thumb.win) > maxSize || WIN_H (ts->thumb.win) > maxSize)
+    if (!ts->thumb.win || !ts->dock)
     {
-	if (WIN_W (ts->thumb.win) >= WIN_H (ts->thumb.win) )
-	    scale = maxSize / WIN_W (ts->thumb.win);
-	else
-	    scale = maxSize / WIN_H (ts->thumb.win);
+	ts->thumb.win  = NULL;
+        ts->thumb.dock = NULL;
+	return;
     }
 
-    ts->thumb.width  = WIN_W (ts->thumb.win) * scale;
-    ts->thumb.height = WIN_H (ts->thumb.win) * scale;
+    w = ts->thumb.win;
+
+    /* do we nee to scale the window down? */
+    if (WIN_W (w) > maxSize || WIN_H (w) > maxSize)
+    {
+	if (WIN_W (w) >= WIN_H (w))
+	    scale = maxSize / WIN_W (w);
+	else
+	    scale = maxSize / WIN_H (w);
+    }
+
+    ts->thumb.width  = WIN_W (w)* scale;
+    ts->thumb.height = WIN_H (w) * scale;
     ts->thumb.scale  = scale;
 
-    THUMB_WINDOW (ts->thumb.win);
-
-    if (thumbnailGetTitleEnabled (s) )
+    if (thumbnailGetTitleEnabled (s))
 	renderThumbText (s, &ts->thumb, FALSE);
     else
 	freeThumbText (s, &ts->thumb);
 
-    int igMidPoint[2] = { tw->ig.x + (tw->ig.width / 2),
-			  tw->ig.y + (tw->ig.height / 2) };
-    int tMidPoint[2];
-    int tPos[2];
-    int tmpPos[2];
+    igMidPoint[0] = w->iconGeometry.x + (w->iconGeometry.width / 2);
+    igMidPoint[1] = w->iconGeometry.y + (w->iconGeometry.height / 2);
 
-    float distance = 1000000;
-
-    int off = thumbnailGetBorder (s);
-
-    int oDev = outputDeviceForPoint (s, tw->ig.x + (tw->ig.width / 2),
-				     tw->ig.y + (tw->ig.height / 2) );
-
-    int ox1, oy1, ox2, oy2, ow, oh;
+    off = thumbnailGetBorder (s);
+    oDev = outputDeviceForPoint (s,
+				 w->iconGeometry.x +
+				 (w->iconGeometry.width / 2),
+				 w->iconGeometry.y +
+				 (w->iconGeometry.height / 2));
 
     if (s->nOutputDev == 1 || oDev > s->nOutputDev)
     {
@@ -352,22 +294,17 @@ thumbUpdateThumbnail (CompScreen *s)
 	oh  = oy2 - oy1;
     }
 
-    int tHeight = ts->thumb.height;
-
+    tHeight = ts->thumb.height;
     if (ts->thumb.textPixmap)
 	tHeight += ts->thumb.tHeight + TEXT_DISTANCE;
 
     // failsave position
     tPos[0] = igMidPoint[0] - (ts->thumb.width / 2.0);
 
-    if (tw->ig.y - tHeight >= 0)
-    {
-	tPos[1] = tw->ig.y - tHeight;
-    }
+    if (w->iconGeometry.y - tHeight >= 0)
+	tPos[1] = w->iconGeometry.y - tHeight;
     else
-    {
-	tPos[1] = tw->ig.y + tw->ig.height;
-    }
+	tPos[1] = w->iconGeometry.y + w->iconGeometry.height;
 
     // above
     tmpPos[0] = igMidPoint[0] - (ts->thumb.width / 2.0);
@@ -459,20 +396,25 @@ static Bool
 thumbShowThumbnail (void *vs)
 {
     CompScreen *s = (CompScreen *) vs;
+
     THUMB_SCREEN (s);
-    ts->showingThumb = TRUE;
+
+    ts->showingThumb   = TRUE;
+    ts->displayTimeout = 0;
+
     thumbUpdateThumbnail (s);
     damageThumbRegion (s, &ts->thumb);
-    return TRUE;
+
+    return FALSE;
 }
 
 static Bool
 checkPosition (CompWindow *w)
 {
-    if (thumbnailGetCurrentViewport (w->screen) )
+    if (thumbnailGetCurrentViewport (w->screen))
     {
 	/* TODO: We need a faster calculation here */
-	Bool onViewport = FALSE;
+	Bool   onViewport = FALSE;
 	Region reg = XCreateRegion ();
 	
 	if (!reg)
@@ -492,120 +434,78 @@ checkPosition (CompWindow *w)
     return TRUE;
 }
 
-static Bool
-thumbUpdateMouse (void *vs)
+static void
+positionUpdate (CompScreen *s,
+		int        x,
+		int        y)
 {
-    CompScreen *s = (CompScreen *) vs;
+    CompWindow *cw    = s->windows;
+    CompWindow *found = NULL;
 
     THUMB_SCREEN (s);
 
-    int          winX, winY;
-    int          rootX, rootY;
-    unsigned int mask_return;
-    Window       root_return;
-    Window       child_return;
-
-    XQueryPointer (s->display->display, s->root,
-		   &root_return, &child_return,
-		   &rootX, &rootY, &winX, &winY, &mask_return);
-
-    CompWindow *w = findWindowAtDisplay (s->display, child_return);
-
-    if (w && w->type & CompWindowTypeDockMask)
+    for (; cw && !found; cw = cw->next)
     {
-	if (ts->dock != w)
-	{
-	    ts->dock = w;
+	if (!cw->iconGeometrySet)
+	    continue;
 
+	if (cw->attrib.map_state != IsViewable)
+	    continue;
+	
+	if (cw->state & CompWindowStateSkipTaskbarMask)
+	    continue;
+
+	if (cw->state & CompWindowStateSkipPagerMask)
+	    continue;
+
+	if (!cw->managed)
+	    continue;
+
+	if (!cw->texture->pixmap)
+	    continue;
+
+	if (x >= cw->iconGeometry.x                          &&
+	    x < cw->iconGeometry.x + cw->iconGeometry.width  &&
+	    y >= cw->iconGeometry.y                          &&
+	    y < cw->iconGeometry.y + cw->iconGeometry.height &&
+	    checkPosition (cw))
+	{
+	    found = cw;
+
+	}
+    }
+
+    if (found)
+    {
+	if (!ts->showingThumb &&
+	    !(ts->thumb.opacity != 0.0 && ts->thumb.win == found))
+	{
 	    if (ts->displayTimeout)
+
 	    {
-		compRemoveTimeout (ts->displayTimeout);
-		ts->displayTimeout = 0;
-	    }
-
-	    ts->pointedWin   = NULL;
-	    ts->showingThumb = FALSE;
-	}
-
-	CompWindow *cw    = s->windows;
-	CompWindow *found = NULL;
-
-	for (; cw && !found; cw = cw->next)
-	{
-	    THUMB_WINDOW (cw);
-
-	    if (!tw->ig.isSet)
-		continue;
-
-	    if (cw->attrib.map_state != IsViewable)
-		continue;
-
-	    if (cw->state & CompWindowStateSkipTaskbarMask)
-		continue;
-
-	    if (cw->state & CompWindowStateSkipPagerMask)
-		continue;
-
-	    if (!cw->managed)
-		continue;
-
-	    if (!w->texture->pixmap)
-		continue;
-
-	    if (rootX >= tw->ig.x && rootX < tw->ig.x + tw->ig.width &&
-		rootY >= tw->ig.y && rootY < tw->ig.y + tw->ig.height &&
-		checkPosition (cw) )
-	    {
-		found = cw;
-	    }
-	}
-
-	if (found)
-	{
-	    if (!ts->showingThumb &&
-		! (ts->thumb.opacity != 0.0 && ts->thumb.win == found) )
-	    {
-		if (ts->displayTimeout)
-		{
-		    if (ts->pointedWin != found)
-		    {
-			compRemoveTimeout (ts->displayTimeout);
-			ts->displayTimeout =
-			    compAddTimeout (thumbnailGetShowDelay (s),
-					    thumbShowThumbnail, s);
-		    }
-		}
-		else
+		if (ts->pointedWin != found)
 		{
+		    compRemoveTimeout (ts->displayTimeout);
 		    ts->displayTimeout =
-			compAddTimeout (thumbnailGetShowDelay (s),
+		        compAddTimeout (thumbnailGetShowDelay (s),
+					thumbnailGetShowDelay (s) + 500,
 					thumbShowThumbnail, s);
 		}
 	    }
-
-	    ts->pointedWin = found;
-
-	    thumbUpdateThumbnail (s);
-	}
-	else
-	{
-	    ts->dock = NULL;
-
-	    if (ts->displayTimeout)
+	    else
 	    {
-		compRemoveTimeout (ts->displayTimeout);
-		ts->displayTimeout = 0;
+		ts->displayTimeout =
+		    compAddTimeout (thumbnailGetShowDelay (s),
+				    thumbnailGetShowDelay (s) + 500,
+				    thumbShowThumbnail, s);
 	    }
+        }
 
-	    ts->pointedWin   = 0;
-	    ts->showingThumb = FALSE;
-	}
-
+        ts->pointedWin = found;
+        thumbUpdateThumbnail (s);
     }
     else
     {
-	ts->dock = NULL;
-
 	if (ts->displayTimeout)
 	{
 	    compRemoveTimeout (ts->displayTimeout);
@@ -615,10 +515,6 @@ thumbUpdateMouse (void *vs)
 	ts->pointedWin   = NULL;
 	ts->showingThumb = FALSE;
     }
-
-    ts->mouseTimeout = compAddTimeout (THUMB_MOUSE_UPDATE_SPEED,
-				       thumbUpdateMouse, s);
-    return FALSE;
 }
 
 
@@ -629,13 +525,15 @@ thumbWindowResizeNotify (CompWindow *w,
 			 int        dwidth,
 			 int        dheight)
 {
-    THUMB_SCREEN (w->screen);
+    CompScreen *s = w->screen;
 
-    thumbUpdateThumbnail (w->screen);
+    THUMB_SCREEN (s);
 
-    UNWRAP (ts, w->screen, windowResizeNotify);
-    (*w->screen->windowResizeNotify) (w, dx, dy, dwidth, dheight);
-    WRAP (ts, w->screen, windowResizeNotify, thumbWindowResizeNotify);
+    thumbUpdateThumbnail (s);
+
+    UNWRAP (ts, s, windowResizeNotify);
+    (*s->windowResizeNotify) (w, dx, dy, dwidth, dheight);
+    WRAP (ts, s, windowResizeNotify, thumbWindowResizeNotify);
 }
 
 static void
@@ -653,16 +551,7 @@ thumbHandleEvent (CompDisplay * d,
     switch (event->type)
     {
     case PropertyNotify:
-	if (event->xproperty.atom == td->winIconGeometryAtom)
-	{
-	    w = findWindowAtDisplay (d, event->xproperty.window);
-
-	    if (w)
-	    {
-		updateWindowIconGeometry (w);
-	    }
-	}
-	else if (event->xproperty.atom == d->wmNameAtom)
+	if (event->xproperty.atom == d->wmNameAtom)
 	{
 	    w = findWindowAtDisplay (d, event->xproperty.window);
 
@@ -685,8 +574,6 @@ thumbHandleEvent (CompDisplay * d,
 
 	    THUMB_SCREEN (s);
 
-	    ts->dock = NULL;
-
 	    if (ts->displayTimeout)
 	    {
 		compRemoveTimeout (ts->displayTimeout);
@@ -698,6 +585,88 @@ thumbHandleEvent (CompDisplay * d,
 	}
 	break;
 
+    case EnterNotify:
+	w = findWindowAtDisplay (d, event->xcrossing.window);
+	if (w)
+	{
+	    CompScreen *s = w->screen;
+
+	    THUMB_SCREEN (s);
+	    THUMB_DISPLAY (w->screen->display);
+
+	    if (w->type & CompWindowTypeDockMask)
+	    {
+		if (ts->dock != w)
+		{
+		    ts->dock = w;
+
+		    if (ts->displayTimeout)
+		    {
+			compRemoveTimeout (ts->displayTimeout);
+			ts->displayTimeout = 0;
+		    }
+
+		    ts->pointedWin   = NULL;
+		    ts->showingThumb = FALSE;
+		}
+
+		if (!ts->pollHandle)
+		{
+		    ts->pollHandle =
+			(*td->mpFunc->addPositionPolling) (s, positionUpdate);
+		}
+	    }
+	    else
+	    {
+		ts->dock = NULL;
+
+		if (ts->displayTimeout)
+		{
+		    compRemoveTimeout (ts->displayTimeout);
+		    ts->displayTimeout = 0;
+		}
+
+		ts->pointedWin   = NULL;
+		ts->showingThumb = FALSE;
+
+		if (ts->pollHandle)
+		{
+		    (*td->mpFunc->removePositionPolling) (s, ts->pollHandle);
+		    ts->pollHandle = 0;
+		}
+	    }
+	}
+	break;
+    case LeaveNotify:
+	w = findWindowAtDisplay (d, event->xcrossing.window);
+	if (w)
+	{
+	    THUMB_SCREEN (w->screen);
+	    THUMB_DISPLAY (w->screen->display);
+
+	    if (w->type & CompWindowTypeDockMask)
+	    {
+		ts->dock = NULL;
+
+		if (ts->displayTimeout)
+		{
+		    compRemoveTimeout (ts->displayTimeout);
+		    ts->displayTimeout = 0;
+		}
+
+		ts->pointedWin   = NULL;
+		ts->showingThumb = FALSE;
+
+		if (ts->pollHandle)
+		{
+		    (*td->mpFunc->removePositionPolling) (w->screen,
+		 					  ts->pollHandle);
+		    ts->pollHandle = 0;
+		}
+	    }
+	}
+	break;
+
     default:
 	break;
     }
@@ -709,20 +678,22 @@ thumbPaintThumb (CompScreen          *s,
 		 Thumbnail           *t,
 		 const CompTransform *transform)
 {
-    THUMB_SCREEN (s);
     AddWindowGeometryProc oldAddWindowGeometry;
-    CompWindow *w = t->win;
+    CompWindow            *w = t->win;
+    int                   wx = t->x;
+    int                   wy = t->y;
+    float                 width  = t->width;
+    float                 height = t->height;
+    WindowPaintAttrib     sAttrib;
+    unsigned int          mask = PAINT_WINDOW_TRANSFORMED_MASK |
+	                         PAINT_WINDOW_TRANSLUCENT_MASK;
+
+    THUMB_SCREEN (s);
 
     if (!w)
 	return;
 
-    int wx = t->x;
-    int wy = t->y;
-
-    float width  = t->width;
-    float height = t->height;
-
-    WindowPaintAttrib sAttrib = w->paint;
+    sAttrib = w->paint;
 
     if (t->textPixmap)
 	height += t->tHeight + TEXT_DISTANCE;
@@ -730,23 +701,21 @@ thumbPaintThumb (CompScreen          *s,
     /* Wrap drawWindowGeometry to make sure the general
        drawWindowGeometry function is used */
     oldAddWindowGeometry = w->screen->addWindowGeometry;
-
     w->screen->addWindowGeometry = addWindowGeometry;
 
-    unsigned int mask = PAINT_WINDOW_TRANSFORMED_MASK |
-			PAINT_WINDOW_TRANSLUCENT_MASK;
-
     if (w->texture->pixmap)
     {
-
-	int off = t->offset;
+	int            off = t->offset;
+	GLenum         filter = s->display->textureFilter;
+	FragmentAttrib fragment;
+	CompTransform  wTransform = *transform;
 
 	glEnable (GL_BLEND);
 	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 	glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
 	glDisableClientState (GL_TEXTURE_COORD_ARRAY);
 	
-	if (thumbnailGetWindowLike (s) )
+	if (thumbnailGetWindowLike (s))
 	{
 	    glColor4f (1.0, 1.0, 1.0, t->opacity);
 	    enableTexture (s, &ts->windowTexture, COMP_TEXTURE_FILTER_GOOD);
@@ -843,7 +812,7 @@ thumbPaintThumb (CompScreen          *s,
 
 	glEnd ();
 
-	if (thumbnailGetWindowLike (s) )
+	if (thumbnailGetWindowLike (s))
 	{
 	    disableTexture (s, &ts->windowTexture);
 	}
@@ -858,6 +827,9 @@ thumbPaintThumb (CompScreen          *s,
 
 	if (t->textPixmap)
 	{
+	    float ox = 0.0;
+	    float w,h;
+
 	    glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
 	    glColor4f (1.0, 1.0, 1.0, t->opacity);
 
@@ -865,23 +837,20 @@ thumbPaintThumb (CompScreen          *s,
 
 	    CompMatrix *m = &t->textTexture.matrix;
 
-	    float ox = 0.0;
-
 	    if (t->tWidth < width)
 		ox = (width - t->tWidth) / 2.0;
 
-	    float w = MIN (width, t->tWidth);
-
-	    float h = t->tHeight;
+	    w = MIN (width, t->tWidth);
+	    h = t->tHeight;
 
 	    glBegin (GL_QUADS);
-	    glTexCoord2f (COMP_TEX_COORD_X (m, 0), COMP_TEX_COORD_Y (m, 0) );
+	    glTexCoord2f (COMP_TEX_COORD_X (m, 0), COMP_TEX_COORD_Y (m, 0));
 	    glVertex2f (wx + ox, wy + height - h);
-	    glTexCoord2f (COMP_TEX_COORD_X (m, 0), COMP_TEX_COORD_Y (m, h) );
+	    glTexCoord2f (COMP_TEX_COORD_X (m, 0), COMP_TEX_COORD_Y (m, h));
 	    glVertex2f (wx + ox, wy + height);
-	    glTexCoord2f (COMP_TEX_COORD_X (m, w), COMP_TEX_COORD_Y (m, h) );
+	    glTexCoord2f (COMP_TEX_COORD_X (m, w), COMP_TEX_COORD_Y (m, h));
 	    glVertex2f (wx + ox + w, wy + height);
-	    glTexCoord2f (COMP_TEX_COORD_X (m, w), COMP_TEX_COORD_Y (m, 0) );
+	    glTexCoord2f (COMP_TEX_COORD_X (m, w), COMP_TEX_COORD_Y (m, 0));
 	    glVertex2f (wx + ox + w, wy + height - h);
 	    glEnd ();
 
@@ -904,15 +873,9 @@ thumbPaintThumb (CompScreen          *s,
 	sAttrib.xTranslate = wx - w->attrib.x + w->input.left * sAttrib.xScale;
 	sAttrib.yTranslate = wy - w->attrib.y + w->input.top * sAttrib.yScale;
 
-	GLenum filter = s->display->textureFilter;
-
-	if (thumbnailGetMipmap (s) )
+	if (thumbnailGetMipmap (s))
 	    s->display->textureFilter = GL_LINEAR_MIPMAP_LINEAR;
 
-	FragmentAttrib fragment;
-
-	CompTransform wTransform = *transform;
-
 	initFragmentAttrib (&fragment, &sAttrib);
 
 	matrixTranslate (&wTransform, w->attrib.x, w->attrib.y, 0.0f);
@@ -924,8 +887,8 @@ thumbPaintThumb (CompScreen          *s,
 
 	glPushMatrix ();
 	glLoadMatrixf (wTransform.m);
-	(w->screen->drawWindow) (w, &wTransform, &fragment, &infiniteRegion,
-				 mask);
+	(*w->screen->drawWindow) (w, &wTransform, &fragment,
+				  &infiniteRegion, mask);
 	glPopMatrix ();
 
 	s->display->textureFilter = filter;
@@ -938,13 +901,14 @@ static void
 thumbPreparePaintScreen (CompScreen *s,
 			 int        ms)
 {
+    float val = ms;
+
     THUMB_SCREEN (s);
 
-    float val = ms;
     val /= 1000;
     val /= thumbnailGetFadeSpeed (s);
 
-    if (otherScreenGrabExist (s, 0) )
+    if (s->maxGrab)
     {
 	ts->dock = NULL;
 
@@ -970,9 +934,15 @@ thumbPreparePaintScreen (CompScreen *s,
 	    ts->thumb.win = NULL;
     }
 
-    ts->oldThumb.opacity = MAX (0.0, ts->oldThumb.opacity - val);
-    if (ts->oldThumb.opacity == 0.0)
+    if (ts->oldThumb.opacity > 0.0f)
+    {
+	ts->oldThumb.opacity = MAX (0.0, ts->oldThumb.opacity - val);
+	if (ts->oldThumb.opacity == 0.0)
+	{
+	    damageThumbRegion (s, &ts->oldThumb);
 	    ts->oldThumb.win = NULL;
+	}
+    }
 
     UNWRAP (ts, s, preparePaintScreen);
     (*s->preparePaintScreen) (s, ms);
@@ -1003,7 +973,8 @@ thumbPaintOutput (CompScreen              *s,
 		  CompOutput              *output,
 		  unsigned int            mask)
 {
-    Bool status;
+    Bool         status;
+    unsigned int newMask = mask;
 
     THUMB_SCREEN (s);
 
@@ -1012,11 +983,11 @@ thumbPaintOutput (CompScreen              *s,
     ts->x = s->x;
     ts->y = s->y;
 
-    unsigned int newMask = mask;
-
-    if ( (ts->oldThumb.opacity > 0.0 && ts->oldThumb.win) ||
-	 (ts->thumb.opacity > 0.0 && ts->thumb.win) )
+    if ((ts->oldThumb.opacity > 0.0 && ts->oldThumb.win) ||
+       	(ts->thumb.opacity > 0.0 && ts->thumb.win))
+    {
 	newMask |= PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK;
+    }
 
     UNWRAP (ts, s, paintOutput);
     status = (*s->paintOutput) (s, sAttrib, transform, region,
@@ -1076,7 +1047,7 @@ thumbPaintTransformedOutput (CompScreen              *s,
 	{
 	    CompTransform sTransform = *transform;
 
-	    (s->applyScreenTransform) (s, sAttrib, output, &sTransform);
+	    (*s->applyScreenTransform) (s, sAttrib, output, &sTransform);
 	    transformToScreenSpace (s, output, -sAttrib->zTranslate,
 				    &sTransform);
 	    glPushMatrix ();
@@ -1089,7 +1060,7 @@ thumbPaintTransformedOutput (CompScreen              *s,
 	{
 	    CompTransform sTransform = *transform;
 
-	    (s->applyScreenTransform) (s, sAttrib, output, &sTransform);
+	    (*s->applyScreenTransform) (s, sAttrib, output, &sTransform);
 	    transformToScreenSpace (s, output, -sAttrib->zTranslate,
 				    &sTransform);
 	    glPushMatrix ();
@@ -1160,9 +1131,16 @@ thumbInitDisplay (CompPlugin  *p,
 		  CompDisplay *d)
 {
     ThumbDisplay *td;
+    int          index;
 
-    td = malloc (sizeof (ThumbDisplay) );
+    if (!checkPluginABI ("core", CORE_ABIVERSION) ||
+	!checkPluginABI ("mousepoll", MOUSEPOLL_ABIVERSION))
+	return FALSE;
 
+    if (!getPluginDisplayIndex (d, "mousepoll", &index))
+	return FALSE;
+
+    td = malloc (sizeof (ThumbDisplay));
     if (!td)
 	return FALSE;
 
@@ -1174,12 +1152,16 @@ thumbInitDisplay (CompPlugin  *p,
 	return FALSE;
     }
 
-    td->winIconGeometryAtom = XInternAtom (d->display,
-					   "_NET_WM_ICON_GEOMETRY", FALSE);
+    td->textAvailable = checkPluginABI ("text", TEXT_ABIVERSION);
+    if (!td->textAvailable)
+	compLogMessage ("thumbnail", CompLogLevelWarn,
+			"No compatible text plugin found.");
+
+    td->mpFunc = d->base.privates[index].ptr;
 
     WRAP (td, d, handleEvent, thumbHandleEvent);
 
-    d->privates[displayPrivateIndex].ptr = td;
+    d->base.privates[displayPrivateIndex].ptr = td;
 
     return TRUE;
 }
@@ -1198,32 +1180,10 @@ thumbFiniDisplay (CompPlugin  *p,
     free (td);
 }
 
-static Bool
-thumbInitWindow (CompPlugin *p,
-		 CompWindow *w)
-{
-    ThumbWindow *tw;
-
-    THUMB_SCREEN (w->screen);
-
-    /* create window */
-    tw = calloc (1, sizeof (ThumbWindow) );
-
-    if (!tw)
-	return FALSE;
-
-    w->privates[ts->windowPrivateIndex].ptr = tw;
-
-    updateWindowIconGeometry (w);
-
-    return TRUE;
-}
-
 static void
 thumbFiniWindow (CompPlugin *p,
 		 CompWindow *w)
 {
-    THUMB_WINDOW (w);
     THUMB_SCREEN (w->screen);
 
     if (ts->thumb.win == w)
@@ -1239,73 +1199,6 @@ thumbFiniWindow (CompPlugin *p,
 	ts->oldThumb.win = NULL;
 	ts->oldThumb.opacity = 0;
     }
-
-    /* free window pointer */
-    free (tw);
-}
-
-static Bool
-thumbRGBAimageToTexture (CompScreen   *screen,
-			 CompTexture  *texture,
-			 char         *image,
-			 unsigned int width,
-			 unsigned int height)
-{
-    char *data;
-    int i;
-
-    makeScreenCurrent (screen);
-
-    data = malloc (4 * width * height);
-
-    if (!data)
-	return FALSE;
-
-    for (i = 0; i < height; i++)
-	memcpy (&data[i * width * 4],
-		&image[ (height - i - 1) * width * 4], width * 4);
-
-    releasePixmapFromTexture (screen, texture);
-
-    if (screen->textureNonPowerOfTwo ||
-	(POWER_OF_TWO (width) && POWER_OF_TWO (height) ) )
-    {
-	texture->target = GL_TEXTURE_2D;
-	texture->matrix.xx = 1.0f / width;
-	texture->matrix.yy = -1.0f / height;
-	texture->matrix.y0 = 1.0f;
-    }
-    else
-    {
-	texture->target = GL_TEXTURE_RECTANGLE_NV;
-	texture->matrix.xx = 1.0f;
-	texture->matrix.yy = -1.0f;
-	texture->matrix.y0 = height;
-    }
-
-    if (!texture->name)
-	glGenTextures (1, &texture->name);
-
-    glBindTexture (texture->target, texture->name);
-
-    glTexImage2D (texture->target, 0, GL_RGBA, width, height, 0, GL_RGBA,
-		  GL_UNSIGNED_BYTE, data);
-
-    texture->filter = GL_NEAREST;
-
-    glTexParameteri (texture->target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
-    glTexParameteri (texture->target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
-    glTexParameteri (texture->target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
-    glTexParameteri (texture->target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
-
-    texture->wrap   = GL_CLAMP_TO_EDGE;
-    texture->mipmap = TRUE;
-
-    glBindTexture (texture->target, 0);
-
-    free (data);
-
-    return TRUE;
 }
 
 static Bool
@@ -1316,13 +1209,10 @@ thumbInitScreen (CompPlugin *p,
 
     THUMB_DISPLAY (s->display);
 
-    ts = calloc (1, sizeof (ThumbScreen) );
-
+    ts = calloc (1, sizeof (ThumbScreen));
     if (!ts)
 	return FALSE;
 
-    ts->windowPrivateIndex = allocateWindowPrivateIndex (s);
-
     WRAP (ts, s, paintOutput, thumbPaintOutput);
     WRAP (ts, s, damageWindowRect, thumbDamageWindowRect);
     WRAP (ts, s, preparePaintScreen, thumbPreparePaintScreen);
@@ -1337,17 +1227,17 @@ thumbInitScreen (CompPlugin *p,
     ts->thumb.win      = NULL;
     ts->oldThumb.win   = NULL;
     ts->showingThumb   = FALSE;
+    ts->pollHandle     = 0;
 
-    s->privates[td->screenPrivateIndex].ptr = ts;
-
-    ts->mouseTimeout =
-	compAddTimeout (THUMB_MOUSE_UPDATE_SPEED, thumbUpdateMouse, s);
+    s->base.privates[td->screenPrivateIndex].ptr = ts;
 
     initTexture (s, &ts->glowTexture);
     initTexture (s, &ts->windowTexture);
 
-    thumbRGBAimageToTexture (s, &ts->glowTexture, glowTex, 32, 32);
-    thumbRGBAimageToTexture (s, &ts->windowTexture, windowTex, 32, 32);
+    imageDataToTexture (s, &ts->glowTexture, glowTex, 32, 32,
+			GL_RGBA, GL_UNSIGNED_BYTE);
+    imageDataToTexture (s, &ts->windowTexture, windowTex, 32, 32,
+			GL_RGBA, GL_UNSIGNED_BYTE);
 
     ts->thumb.textPixmap    = None;
     ts->oldThumb.textPixmap = None;
@@ -1370,8 +1260,13 @@ thumbFiniScreen (CompPlugin *p,
     UNWRAP (ts, s, windowResizeNotify);
     UNWRAP (ts, s, paintTransformedOutput);
 
-    if (ts->mouseTimeout)
-	compRemoveTimeout (ts->mouseTimeout);
+    if (ts->pollHandle)
+    {
+	THUMB_DISPLAY (s->display);
+
+	(*td->mpFunc->removePositionPolling) (s, ts->pollHandle);
+	ts->pollHandle = 0;
+    }
 
     freeThumbText (s, &ts->thumb);
     freeThumbText (s, &ts->oldThumb);
@@ -1387,7 +1282,6 @@ static Bool
 thumbInit (CompPlugin *p)
 {
     displayPrivateIndex = allocateDisplayPrivateIndex ();
-
     if (displayPrivateIndex < 0)
 	return FALSE;
 
@@ -1398,33 +1292,43 @@ thumbInit (CompPlugin *p)
 static void
 thumbFini (CompPlugin *p)
 {
-    if (displayPrivateIndex >= 0)
-	freeDisplayPrivateIndex (displayPrivateIndex);
+    freeDisplayPrivateIndex (displayPrivateIndex);
 }
 
-static int
-thumbnailGetVersion (CompPlugin *plugin,
-		     int        version)
+static CompBool
+thumbInitObject (CompPlugin *p,
+		 CompObject *o)
 {
-    return ABIVERSION;
+    static InitPluginObjectProc dispTab[] = {
+	(InitPluginObjectProc) 0, /* InitCore */
+	(InitPluginObjectProc) thumbInitDisplay,
+	(InitPluginObjectProc) thumbInitScreen
+    };
+
+    RETURN_DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), TRUE, (p, o));
 }
 
+static void
+thumbFiniObject (CompPlugin *p,
+		 CompObject *o)
+{
+    static FiniPluginObjectProc dispTab[] = {
+	(FiniPluginObjectProc) 0, /* FiniCore */
+	(FiniPluginObjectProc) thumbFiniDisplay,
+	(FiniPluginObjectProc) thumbFiniScreen,
+	(FiniPluginObjectProc) thumbFiniWindow
+    };
+
+    DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), (p, o));
+}
 
 CompPluginVTable thumbVTable = {
-
     "thumbnail",
-    thumbnailGetVersion,
     0,
     thumbInit,
     thumbFini,
-    thumbInitDisplay,
-    thumbFiniDisplay,
-    thumbInitScreen,
-    thumbFiniScreen,
-    thumbInitWindow,
-    thumbFiniWindow,
-    0,			/* thumbGetDisplayOptions */
-    0,			/* thumbSetDisplayOption */
+    thumbInitObject,
+    thumbFiniObject,
     0,
     0
 };
