@@ -23,54 +23,186 @@
 #include <string.h>
 #include <limits.h>
 
-#include <compiz.h>
+#include <compiz-core.h>
 #include <X11/Xatom.h>
-#include <workarounds_options.h>
+#include "workarounds_options.h"
 
 static int displayPrivateIndex;
 
+typedef struct _WorkaroundsManagedFsWindow {
+	Window id;
+	struct _WorkaroundsManagedFsWindow *next;
+} WorkaroundsManagedFsWindow;
+
 typedef struct _WorkaroundsDisplay {
     int screenPrivateIndex;
 
     HandleEventProc handleEvent;
 
-    Atom roleAtom;
+    Atom                       roleAtom;
+    WorkaroundsManagedFsWindow *mfwList;
 } WorkaroundsDisplay;
 
+typedef void (*GLProgramParameter4dvProc) (GLenum         target,
+					   GLuint         index,
+					   const GLdouble *data);
+
 typedef struct _WorkaroundsScreen {
     int windowPrivateIndex;
 
     WindowResizeNotifyProc         windowResizeNotify;
     GetAllowedActionsForWindowProc getAllowedActionsForWindow;
+    PaintScreenProc                paintScreen;
+
+    GLProgramParameter4fProc       origProgramEnvParameter4f;
+    GLProgramParameter4dvProc      programEnvParameter4dv;
+
+    GLXCopySubBufferProc           origCopySubBuffer;
 } WorkaroundsScreen;
 
+CompScreen *currentScreen = NULL;
+
 typedef struct _WorkaroundsWindow {
     Bool adjustedWinType;
     Bool madeSticky;
     Bool madeFullscreen;
     Bool isFullscreen;
+    Bool madeDemandAttention;
 } WorkaroundsWindow;
 
 #define GET_WORKAROUNDS_DISPLAY(d) \
-    ((WorkaroundsDisplay *) (d)->privates[displayPrivateIndex].ptr)
+    ((WorkaroundsDisplay *) (d)->base.privates[displayPrivateIndex].ptr)
 
 #define WORKAROUNDS_DISPLAY(d) \
     WorkaroundsDisplay *wd = GET_WORKAROUNDS_DISPLAY (d)
 
 #define GET_WORKAROUNDS_SCREEN(s, wd) \
-    ((WorkaroundsScreen *) (s)->privates[(wd)->screenPrivateIndex].ptr)
+    ((WorkaroundsScreen *) (s)->base.privates[(wd)->screenPrivateIndex].ptr)
 
 #define WORKAROUNDS_SCREEN(s) \
     WorkaroundsScreen *ws = GET_WORKAROUNDS_SCREEN (s, \
                             GET_WORKAROUNDS_DISPLAY (s->display))
 
 #define GET_WORKAROUNDS_WINDOW(w, ns) \
-    ((WorkaroundsWindow *) (w)->privates[(ns)->windowPrivateIndex].ptr)
+    ((WorkaroundsWindow *) (w)->base.privates[(ns)->windowPrivateIndex].ptr)
 #define WORKAROUNDS_WINDOW(w) \
     WorkaroundsWindow *ww = GET_WORKAROUNDS_WINDOW  (w, \
 		    GET_WORKAROUNDS_SCREEN  (w->screen, \
 		    GET_WORKAROUNDS_DISPLAY (w->screen->display)))
 
+static void
+workaroundsAddToFullscreenList (CompWindow *w)
+{
+    WorkaroundsManagedFsWindow *mfw, *nmfw;
+
+    WORKAROUNDS_DISPLAY (w->screen->display);
+
+    nmfw = malloc (sizeof (WorkaroundsManagedFsWindow));
+    if (!nmfw)
+	return;
+
+    nmfw->id   = w->id;
+    nmfw->next = NULL;
+
+    if (!wd->mfwList)
+	wd->mfwList = nmfw;
+    else
+    {
+	for (mfw = wd->mfwList; mfw->next; mfw = mfw->next)
+	    if (mfw->id == w->id)
+	    {
+		free (nmfw);
+		return;
+	    }
+
+	mfw->next = nmfw;
+    }
+}
+
+static void
+workaroundsRemoveFromFullscreenList (CompWindow *w)
+{
+    WorkaroundsManagedFsWindow *mfw;
+    WorkaroundsManagedFsWindow *temp;
+
+    WORKAROUNDS_DISPLAY (w->screen->display);
+
+    if (!wd->mfwList)
+	return;
+
+    if (wd->mfwList->id == w->id)
+    {
+	mfw = wd->mfwList;
+	wd->mfwList = wd->mfwList->next;
+	free (mfw);
+	return;
+    }
+
+    for (mfw = wd->mfwList; mfw->next; mfw = mfw->next) 
+    {
+	if (mfw->next->id == w->id) 
+	{
+	    temp = mfw->next;
+	    mfw->next = mfw->next->next;
+	    free (temp);
+	    break;
+	}
+    }
+}
+
+static void
+workaroundsProgramEnvParameter4f (GLenum  target,
+				  GLuint  index,
+				  GLfloat x,
+				  GLfloat y,
+				  GLfloat z,
+				  GLfloat w)
+{
+    WorkaroundsScreen *ws;
+    GLdouble data[4];
+
+    if (!currentScreen)
+	return;
+
+    ws = GET_WORKAROUNDS_SCREEN (currentScreen, GET_WORKAROUNDS_DISPLAY (
+				 currentScreen->display));
+
+    data[0] = x;
+    data[1] = y;
+    data[2] = z;
+    data[3] = w;
+
+    (*ws->programEnvParameter4dv) (target, index, data);
+}
+
+static void
+workaroundsUpdateParameterFix (CompScreen *s)
+{
+    WORKAROUNDS_SCREEN (s);
+
+    if (!s->programEnvParameter4f || !ws->programEnvParameter4dv)
+	return;
+    if (workaroundsGetAiglxFragmentFix (s->display))
+	s->programEnvParameter4f = workaroundsProgramEnvParameter4f;
+    else
+	s->programEnvParameter4f = ws->origProgramEnvParameter4f;
+}
+
+static void
+workaroundsPaintScreen (CompScreen   *s,
+			CompOutput   *outputs,
+			int          numOutputs,
+			unsigned int mask)
+{
+    WORKAROUNDS_SCREEN (s);
+
+    currentScreen = s;
+
+    UNWRAP (ws, s, paintScreen);
+    (*s->paintScreen) (s, outputs, numOutputs, mask);
+    WRAP (ws, s, paintScreen, workaroundsPaintScreen);
+}
+
 static char *
 workaroundsGetWindowRoleAtom (CompWindow *w)
 {
@@ -139,6 +271,35 @@ workaroundsUpdateSticky (CompWindow *w)
 	workaroundsRemoveSticky (w);
 }
 
+static void
+updateUrgencyState (CompWindow *w)
+{
+    XWMHints *hints;
+    Bool     urgent = FALSE;
+
+    WORKAROUNDS_WINDOW (w);
+
+    hints = XGetWMHints (w->screen->display->display, w->id);
+    if (hints)
+    {
+	if (hints->flags & XUrgencyHint)
+	    urgent = TRUE;
+
+	XFree (hints);
+    }
+
+    if (urgent)
+    {
+	ww->madeDemandAttention = TRUE;
+	changeWindowState (w, w->state | CompWindowStateDemandsAttentionMask);
+    }
+    else if (ww->madeDemandAttention)
+    {
+	ww->madeDemandAttention = FALSE;
+	changeWindowState (w, w->state & ~CompWindowStateDemandsAttentionMask);
+    }
+}
+
 static void
 workaroundsGetAllowedActionsForWindow (CompWindow   *w,
 				       unsigned int *setActions,
@@ -166,6 +327,10 @@ workaroundsFixupFullscreen (CompWindow *w)
     BoxPtr box;
 
     WORKAROUNDS_WINDOW (w);
+    WORKAROUNDS_DISPLAY (w->screen->display);
+
+    if (!workaroundsGetLegacyFullscreen (w->screen->display))
+	return;
 
     if (w->wmType & CompWindowTypeDesktopMask)
     {
@@ -196,8 +361,7 @@ workaroundsFixupFullscreen (CompWindow *w)
     }
 
     ww->isFullscreen = isFullSize;
-    if ((isFullSize && !(w->state & CompWindowStateFullscreenMask)) ||
-	(!isFullSize && (w->state & CompWindowStateFullscreenMask)))
+    if (isFullSize && !(w->state & CompWindowStateFullscreenMask))
     {
 	unsigned int state = w->state & ~CompWindowStateFullscreenMask;
 
@@ -208,11 +372,40 @@ workaroundsFixupFullscreen (CompWindow *w)
 	if (state != w->state)
 	{
 	    changeWindowState (w, state);
-	    recalcWindowType (w);
-	    recalcWindowActions (w);
 	    updateWindowAttributes (w, CompStackingUpdateModeNormal);
+
+	    /* keep track of windows that we interact with */
+	    workaroundsAddToFullscreenList (w);
 	}
     }
+    else if (!isFullSize && wd->mfwList &&
+	     (w->state & CompWindowStateFullscreenMask))
+    {
+	/* did we set the flag? */
+	WorkaroundsManagedFsWindow *mfw;
+
+	for (mfw = wd->mfwList; mfw->next; mfw = mfw->next) 
+	{
+	    if (mfw->id == w->id) 
+	    {
+		unsigned int state = w->state & ~CompWindowStateFullscreenMask;
+
+		if (isFullSize)
+		    state |= CompWindowStateFullscreenMask;
+
+		ww->madeFullscreen = isFullSize;
+
+		if (state != w->state)
+		{
+		    changeWindowState (w, state);
+		    updateWindowAttributes (w, CompStackingUpdateModeNormal);
+		}
+
+		workaroundsRemoveFromFullscreenList (w);
+		break;
+	    }
+    	}
+   }
 }
 
 static void
@@ -226,46 +419,63 @@ workaroundsDoFixes (CompWindow *w)
     /* FIXME: Is this the best way to detect a notification type window? */
     if (workaroundsGetNotificationDaemonFix (d))
     {
-        if (w->wmType == CompWindowTypeNormalMask &&
+        if (newWmType == CompWindowTypeNormalMask &&
             w->attrib.override_redirect && w->resName &&
             strcmp (w->resName, "notification-daemon") == 0)
         {
             newWmType = CompWindowTypeNotificationMask;
+	    goto AppliedFix;
         }
     }
 
-    if ((w->wmType == newWmType) && workaroundsGetFirefoxMenuFix (d))
+    if (workaroundsGetFirefoxMenuFix (d))
     {
-        if (w->wmType == CompWindowTypeNormalMask &&
+        if (newWmType == CompWindowTypeNormalMask &&
             w->attrib.override_redirect && w->resName)
 	{
 	    if ((strcasecmp (w->resName, "gecko") == 0) ||
 		(strcasecmp (w->resName, "popup") == 0))
 	    {
 		newWmType = CompWindowTypeDropdownMenuMask;
+		goto AppliedFix;
 	    }
 	}
     }
 
+    if (workaroundsGetOooMenuFix (d))
+    {
+        if (newWmType == CompWindowTypeNormalMask &&
+            w->attrib.override_redirect && w->resName)
+	{
+	    if (strcasecmp (w->resName, "VCLSalFrame") == 0)
+	    {
+		newWmType = CompWindowTypeDropdownMenuMask;
+		goto AppliedFix;
+	    }
+	}
+    }
     /* FIXME: Basic hack to get Java windows working correctly. */
-    if ((w->wmType == newWmType) && workaroundsGetJavaFix (d) && w->resName)
+    if (workaroundsGetJavaFix (d) && w->resName)
     {
         if ((strcmp (w->resName, "sun-awt-X11-XMenuWindow") == 0) ||
             (strcmp (w->resName, "sun-awt-X11-XWindowPeer") == 0))
         {
             newWmType = CompWindowTypeDropdownMenuMask;
+	    goto AppliedFix;
         }
         else if (strcmp (w->resName, "sun-awt-X11-XDialogPeer") == 0)
         {
             newWmType = CompWindowTypeDialogMask;
+	    goto AppliedFix;
         }
         else if (strcmp (w->resName, "sun-awt-X11-XFramePeer") == 0)
         {
             newWmType = CompWindowTypeNormalMask;
+	    goto AppliedFix;
         }
     }
 
-    if ((newWmType == w->wmType) && workaroundsGetQtFix (d))
+    if (workaroundsGetQtFix (d))
     {
         char *windowRole;
 
@@ -276,24 +486,27 @@ workaroundsDoFixes (CompWindow *w)
             if ((strcmp (windowRole, "toolTipTip") == 0) ||
                 (strcmp (windowRole, "qtooltip_label") == 0))
             {
+		free (windowRole);
                 newWmType = CompWindowTypeTooltipMask;
+		goto AppliedFix;
             }
-
-            free (windowRole);
+	    else
+	    {
+		free (windowRole);
+	    }
         }
 
         /* fix Qt transients - FIXME: is there a better way to detect them? */
-        if (w->wmType == newWmType)
-        {
-            if (!w->resName && w->attrib.override_redirect &&
-		(w->attrib.class == InputOutput) &&
-                (w->wmType == CompWindowTypeUnknownMask))
-            {
-                newWmType = CompWindowTypeDropdownMenuMask;
-            }
-        }
+	if (!w->resName && w->attrib.override_redirect &&
+	    (w->attrib.class == InputOutput) &&
+	    (newWmType == CompWindowTypeUnknownMask))
+	{
+	    newWmType = CompWindowTypeDropdownMenuMask;
+	    goto AppliedFix;
+	}
     }
 
+AppliedFix:
     if (newWmType != w->wmType)
     {
 	WORKAROUNDS_WINDOW (w);
@@ -306,9 +519,6 @@ workaroundsDoFixes (CompWindow *w)
 
 	(*d->matchPropertyChanged) (d, w);
     }
-
-    if (workaroundsGetLegacyFullscreen (w->screen->display))
-	workaroundsFixupFullscreen (w);
 }
 
 static void
@@ -319,9 +529,19 @@ workaroundsDisplayOptionChanged (CompDisplay               *d,
     CompScreen *s;
     CompWindow *w;
 
+    WorkaroundsScreen *ws;
+
     for (s = d->screens; s; s = s->next)
+    {
+	ws = GET_WORKAROUNDS_SCREEN (s, GET_WORKAROUNDS_DISPLAY (d));
 	for (w = s->windows; w; w = w->next)
 	    workaroundsUpdateSticky (w);
+	workaroundsUpdateParameterFix (s);
+	if (workaroundsGetFglrxXglFix (d))
+	    s->copySubBuffer = NULL;
+	else
+	    s->copySubBuffer = ws->origCopySubBuffer;
+    }
 }
 
 static void
@@ -349,12 +569,21 @@ workaroundsHandleEvent (CompDisplay *d,
 	{
 	    workaroundsUpdateSticky (w);
 	    workaroundsDoFixes (w);
+	    workaroundsFixupFullscreen (w);
 	}
 	break;
     case MapNotify:
 	w = findWindowAtDisplay (d, event->xmap.window);
 	if (w && w->attrib.override_redirect)
+	{
 	    workaroundsDoFixes (w);
+	    workaroundsFixupFullscreen (w);
+	}
+	break;
+    case DestroyNotify:
+	w = findWindowAtDisplay (d, event->xdestroywindow.window);
+	if (w)
+	    workaroundsRemoveFromFullscreenList (w);
 	break;
     }
 
@@ -389,6 +618,15 @@ workaroundsHandleEvent (CompDisplay *d,
 	    if (w)
 		workaroundsDoFixes (w);
 	}
+	else if (event->xproperty.atom == XA_WM_HINTS)
+	{
+	    if (workaroundsGetConvertUrgency (d))
+	    {
+		w = findWindowAtDisplay (d, event->xproperty.window);
+		if (w)
+		    updateUrgencyState (w);
+	    }
+	}
 	break;
     default:
 	break;
@@ -401,7 +639,7 @@ workaroundsWindowResizeNotify (CompWindow *w, int dx, int dy,
 {
     WORKAROUNDS_SCREEN (w->screen);
 
-    if (workaroundsGetLegacyFullscreen (w->screen->display))
+    if (w->attrib.map_state == IsViewable)
 	workaroundsFixupFullscreen (w);
 
     UNWRAP (ws, w->screen, windowResizeNotify);
@@ -414,11 +652,14 @@ workaroundsInitDisplay (CompPlugin *plugin, CompDisplay *d)
 {
     WorkaroundsDisplay *wd;
 
+    if (!checkPluginABI ("core", CORE_ABIVERSION))
+	return FALSE;
+
     wd = malloc (sizeof (WorkaroundsDisplay));
     if (!wd)
         return FALSE;
 
-    wd->screenPrivateIndex = allocateScreenPrivateIndex( d );
+    wd->screenPrivateIndex = allocateScreenPrivateIndex (d);
     if (wd->screenPrivateIndex < 0)
     {
         free (wd);
@@ -426,12 +667,16 @@ workaroundsInitDisplay (CompPlugin *plugin, CompDisplay *d)
     }
 
     wd->roleAtom = XInternAtom (d->display, "WM_WINDOW_ROLE", 0);
+    wd->mfwList  = NULL;
 
     workaroundsSetStickyAlldesktopsNotify (d, workaroundsDisplayOptionChanged);
     workaroundsSetAlldesktopStickyMatchNotify (d,
 					       workaroundsDisplayOptionChanged);
+    workaroundsSetAiglxFragmentFixNotify (d, workaroundsDisplayOptionChanged);
+    workaroundsSetFglrxXglFixNotify (d, workaroundsDisplayOptionChanged);
+    
 
-    d->privates[displayPrivateIndex].ptr = wd;
+    d->base.privates[displayPrivateIndex].ptr = wd;
 
     WRAP (wd, d, handleEvent, workaroundsHandleEvent);
 
@@ -468,11 +713,24 @@ workaroundsInitScreen (CompPlugin *plugin, CompScreen *s)
 	return FALSE;
     }
 
+    ws->programEnvParameter4dv = (GLProgramParameter4dvProc)
+	    (*s->getProcAddress) ((GLubyte *) "glProgramEnvParameter4dvARB");
+    ws->origProgramEnvParameter4f = s->programEnvParameter4f;
+
+    ws->origCopySubBuffer = s->copySubBuffer;
+
     WRAP (ws, s, windowResizeNotify, workaroundsWindowResizeNotify);
     WRAP (ws, s, getAllowedActionsForWindow,
 	  workaroundsGetAllowedActionsForWindow);
 
-    s->privates[wd->screenPrivateIndex].ptr = ws;
+    WRAP (ws, s, paintScreen, workaroundsPaintScreen);
+
+    s->base.privates[wd->screenPrivateIndex].ptr = ws;
+
+    workaroundsUpdateParameterFix (s);
+
+    if (workaroundsGetFglrxXglFix (s->display))
+	s->copySubBuffer = NULL;
 
     return TRUE;
 }
@@ -487,15 +745,20 @@ workaroundsFiniScreen (CompPlugin *plugin, CompScreen *s)
     UNWRAP (ws, s, windowResizeNotify);
     UNWRAP (ws, s, getAllowedActionsForWindow);
 
+    UNWRAP (ws, s, paintScreen);
+
+    s->copySubBuffer = ws->origCopySubBuffer;
+
     free (ws);
 }
 
 static Bool
 workaroundsInitWindow (CompPlugin *plugin, CompWindow *w)
 {
-    WORKAROUNDS_SCREEN (w->screen);
     WorkaroundsWindow *ww;
 
+    WORKAROUNDS_SCREEN (w->screen);
+
     ww = malloc (sizeof (WorkaroundsWindow));
     if (!ww)
 	return FALSE;
@@ -504,8 +767,9 @@ workaroundsInitWindow (CompPlugin *plugin, CompWindow *w)
     ww->adjustedWinType = FALSE;
     ww->isFullscreen = FALSE;
     ww->madeFullscreen = FALSE;
+    ww->madeDemandAttention = FALSE;
 
-    w->privates[ws->windowPrivateIndex].ptr = ww;
+    w->base.privates[ws->windowPrivateIndex].ptr = ww;
 
     return TRUE;
 }
@@ -515,20 +779,51 @@ workaroundsFiniWindow (CompPlugin *plugin, CompWindow *w)
 {
     WORKAROUNDS_WINDOW (w);
 
-    if (ww->adjustedWinType)
+    if (!w->destroyed)
     {
-	w->wmType = getWindowType (w->screen->display, w->id);
-	recalcWindowType (w);
-	recalcWindowActions (w);
+	if (ww->adjustedWinType)
+	{
+	    w->wmType = getWindowType (w->screen->display, w->id);
+	    recalcWindowType (w);
+	    recalcWindowActions (w);
+	}
+
+	if (w->state & CompWindowStateStickyMask && ww->madeSticky)
+	    setWindowState (w->screen->display,
+			    w->state & ~CompWindowStateStickyMask, w->id);
     }
 
-     if (w->state & CompWindowStateStickyMask && ww->madeSticky)
-	setWindowState (w->screen->display,
-			w->state & ~CompWindowStateStickyMask, w->id);
-
     free (ww);
 }
 
+static CompBool
+workaroundsInitObject (CompPlugin *p,
+		       CompObject *o)
+{
+    static InitPluginObjectProc dispTab[] = {
+	(InitPluginObjectProc) 0, /* InitCore */
+	(InitPluginObjectProc) workaroundsInitDisplay,
+	(InitPluginObjectProc) workaroundsInitScreen,
+	(InitPluginObjectProc) workaroundsInitWindow
+    };
+
+    RETURN_DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), TRUE, (p, o));
+}
+
+static void
+workaroundsFiniObject (CompPlugin *p,
+		       CompObject *o)
+{
+    static FiniPluginObjectProc dispTab[] = {
+	(FiniPluginObjectProc) 0, /* FiniCore */
+	(FiniPluginObjectProc) workaroundsFiniDisplay,
+	(FiniPluginObjectProc) workaroundsFiniScreen,
+	(FiniPluginObjectProc) workaroundsFiniWindow
+    };
+
+    DISPATCH (o, dispTab, ARRAY_SIZE (dispTab), (p, o));
+}
+
 static Bool
 workaroundsInit (CompPlugin *plugin)
 {
@@ -545,29 +840,16 @@ workaroundsFini (CompPlugin *plugin)
     freeDisplayPrivateIndex (displayPrivateIndex);
 }
 
-static int
-workaroundsGetVersion (CompPlugin *plugin, int version)
-{
-    return ABIVERSION;
-}
-
 CompPluginVTable workaroundsVTable =
 {
     "workarounds",
-    workaroundsGetVersion,
     0,
     workaroundsInit,
     workaroundsFini,
-    workaroundsInitDisplay,
-    workaroundsFiniDisplay,
-    workaroundsInitScreen,
-    workaroundsFiniScreen,
-    workaroundsInitWindow,
-    workaroundsFiniWindow,
-    0, /* GetDisplayOptions */
-    0, /* SetDisplayOption */
-    0, /* GetScreenOptions */
-    0  /* SetScreenOption */
+    workaroundsInitObject,
+    workaroundsFiniObject,
+    0,
+    0
 };
 
 CompPluginVTable *
